/**
 * Define api of apeman-api-watch
 * @function define
 * @param {string} pattern - File name pattern to watch
 * @param {Object} [options] - Optional settings
 * @param {boolean} [options.logging] - Enable logger
 * @param {string} [options.root=process.cwd()] - Root directory path
 * @returns {Object} - API configuration
 */
'use strict'

const co = require('co')
const awatch = require('awatch')
const asleep = require('asleep')
const aslogger = require('aslogger')
const describe = require('./describe')
const uuid = require('uuid')
const apemanApiSSE = require('apeman-api-sse')
const apemanApiFS = require('apeman-api-fs')
const load = require('apeman-api-client/load')

/** @lends define */
function define (options = {}) {
  let logger = aslogger({ disabled: !options.logging })
  let { root = process.cwd() } = options

  let watches = {}
  let clients = {}

  let fsApi = apemanApiFS({ root, logging: false })
  let sseApi = apemanApiSSE((client) => co(function * () {
    let id = uuid.v4()

    clients[ id ] = client
    let closed = false
    client.once('close', () => {
      closed = true
    })
    client.send('start', id)
    while (!closed) {
      yield asleep(10)
    }
    client.send('stop', id)
    delete clients[ id ]
  }), { logging: false })

  return {
    ping (pong = 'pong') {
      return co(function * () {
        logger.trace('pong:', pong)
        return pong
      })
    },

    default () {
      const s = this
      return co(function * () {
        let sse = yield load(sseApi, { $$ctx: s })
        return yield sse.default.call(s)
      })
    },

    start (pattern) {
      const s = this
      return co(function * () {
        let fs = yield load(fsApi, { $$ctx: s })
        logger.debug(`Files to start: ${pattern}`)
        let filenames = yield fs.glob(pattern, {})
        for (let filename of filenames) {
          let watch = yield awatch(filename, (event, filename) => {
            for (let id of Object.keys(clients)) {
              let client = clients[ id ]
              client.send(event, filename)
            }
          })
          Object.assign(watch, { startAt: new Date() })
          let oldWatch = watches[ filename ]
          if (oldWatch) {
            oldWatch.close() // Close old watch
          }
          watches[ filename ] = watch
          logger.trace(`  ${filename}`)
        }
      })
    },

    stop (pattern) {
      const s = this
      return co(function * () {
        let fs = yield load(fsApi, { $$ctx: s })
        logger.debug(`Files to stop: ${pattern}`)
        let filenames = yield fs.glob(pattern, {})
        for (let filename of filenames) {
          let watch = watches[ filename ]
          watch.close()
          delete watches[ filename ]
          logger.trace(`  ${filename}`)
        }
      })
    },

    list () {
      return co(function * () {
        return Object.keys(watches).map((filename) => ({
          filename,
          startAt: watches[ filename ].startAt
        }))
      })
    },

    stopAll () {
      const s = this
      return co(function * () {
        let filenames = Object.keys(watches)
        return yield s.stop(filenames)
      })
    },

    /**
     * A hook which runs before method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @returns {Promise}
     */
    $before (methodName, params) {
      const s = this
      let { state } = s
      return co(function * () {
        state.invokedAt = new Date()
        logger.info(`${methodName} started...`)
      })
    },
    /**
     * A hook which runs after method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @param {?*} returns - Return values
     * @returns {Promise}
     */
    $after (methodName, params, returns) {
      const s = this
      let { state } = s
      return co(function * () {
        let took = new Date() - state.invokedAt
        logger.info(`...${methodName} done! (${took}ms)`)
      })
    },
    /**
     * Method specifications
     */
    $spec: describe()
  }
}

module.exports = define
